@@ -1,13 +1,11 @@
 /*
   +----------------------------------------------------------------------+
-  | PHP Version 5                                                        |
-  +----------------------------------------------------------------------+
-  | Copyright (c) 1997-2016 The PHP Group                                |
+  | Copyright (c) The PHP Group                                          |
   +----------------------------------------------------------------------+
   | This source file is subject to version 3.01 of the PHP license,      |
   | that is bundled with this package in the file LICENSE, and is        |
   | available through the world-wide-web at the following url:           |
-  | http://www.php.net/license/3_01.txt                                  |
+  | https://www.php.net/license/3_01.txt                                 |
   | If you did not receive a copy of the PHP license and are unable to   |
   | obtain it through the world-wide-web, please send a note to          |
   | license@php.net so we can mail you a copy immediately.               |
@@ -16,8 +14,6 @@
   +----------------------------------------------------------------------+
 */
 
-/* $Id$ */
-
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
@@ -36,7 +32,7 @@
 #define STMT_CALL(name, params)											\
 	do {																\
 		S->last_err = name params;										\
-		S->last_err = _oci_error(S->err, stmt->dbh, stmt, #name, S->last_err, FALSE, __FILE__, __LINE__ TSRMLS_CC); \
+		S->last_err = _oci_error(S->err, stmt->dbh, stmt, #name, S->last_err, FALSE, __FILE__, __LINE__); \
 		if (S->last_err) {												\
 			return 0;													\
 		}																\
@@ -45,15 +41,24 @@
 #define STMT_CALL_MSG(name, msg, params)								\
 	do { 																\
 		S->last_err = name params;										\
-		S->last_err = _oci_error(S->err, stmt->dbh, stmt, #name ": " #msg, S->last_err, FALSE, __FILE__, __LINE__ TSRMLS_CC); \
+		S->last_err = _oci_error(S->err, stmt->dbh, stmt, #name ": " #msg, S->last_err, FALSE, __FILE__, __LINE__); \
 		if (S->last_err) {												\
 			return 0;													\
 		}																\
 	} while(0)
 
-static php_stream *oci_create_lob_stream(pdo_stmt_t *stmt, OCILobLocator *lob TSRMLS_DC);
+static php_stream *oci_create_lob_stream(zval *dbh, pdo_stmt_t *stmt, OCILobLocator *lob);
 
-static int oci_stmt_dtor(pdo_stmt_t *stmt TSRMLS_DC) /* {{{ */
+#define OCI_TEMPLOB_CLOSE(envhp, svchp, errhp, lob)				\
+	do															\
+	{															\
+		boolean isTempLOB;										\
+		OCILobIsTemporary(envhp, errhp, lob, &isTempLOB);		\
+		if (isTempLOB)											\
+			OCILobFreeTemporary(svchp, errhp, lob);				\
+	} while(0)
+
+static int oci_stmt_dtor(pdo_stmt_t *stmt) /* {{{ */
 {
 	pdo_oci_stmt *S = (pdo_oci_stmt*)stmt->driver_data;
 	HashTable *BC = stmt->bound_columns;
@@ -93,12 +98,21 @@ static int oci_stmt_dtor(pdo_stmt_t *stmt TSRMLS_DC) /* {{{ */
 		S->einfo.errmsg = NULL;
 	}
 
+	/* TODO: There's php_pdo_stmt_valid_db_obj_handle in PHP-8.5-dev that does these checks. */
+	bool server_obj_usable = !Z_ISUNDEF(stmt->database_object_handle)
+		&& IS_OBJ_VALID(EG(objects_store).object_buckets[Z_OBJ_HANDLE(stmt->database_object_handle)])
+		&& !(OBJ_FLAGS(Z_OBJ(stmt->database_object_handle)) & IS_OBJ_FREE_CALLED);
+
 	if (S->cols) {
 		for (i = 0; i < stmt->column_count; i++) {
 			if (S->cols[i].data) {
 				switch (S->cols[i].dtype) {
 					case SQLT_BLOB:
 					case SQLT_CLOB:
+						if (server_obj_usable) {
+							OCI_TEMPLOB_CLOSE(S->H->env, S->H->svc, S->H->err,
+								(OCILobLocator *) S->cols[i].data);
+						}
 						OCIDescriptorFree(S->cols[i].data, OCI_DTYPE_LOB);
 						break;
 					default:
@@ -116,7 +130,7 @@ static int oci_stmt_dtor(pdo_stmt_t *stmt TSRMLS_DC) /* {{{ */
 	return 1;
 } /* }}} */
 
-static int oci_stmt_execute(pdo_stmt_t *stmt TSRMLS_DC) /* {{{ */
+static int oci_stmt_execute(pdo_stmt_t *stmt) /* {{{ */
 {
 	pdo_oci_stmt *S = (pdo_oci_stmt*)stmt->driver_data;
 	ub4 rowcount;
@@ -188,28 +202,32 @@ static sb4 oci_bind_input_cb(dvoid *ctx, OCIBind *bindp, ub4 iter, ub4 index, dv
 {
 	struct pdo_bound_param_data *param = (struct pdo_bound_param_data*)ctx;
 	pdo_oci_bound_param *P = (pdo_oci_bound_param*)param->driver_data;
-	TSRMLS_FETCH();
+	zval *parameter;
 
-	if (!param || !param->parameter) {
-		php_error_docref(NULL TSRMLS_CC, E_WARNING, "param is NULL in oci_bind_input_cb; this should not happen");
-		return OCI_ERROR;
-	}
+	ZEND_ASSERT(param);
 
 	*indpp = &P->indicator;
 
+	if (Z_ISREF(param->parameter))
+		parameter = Z_REFVAL(param->parameter);
+	else
+		parameter = &param->parameter;
+
 	if (P->thing) {
 		*bufpp = P->thing;
 		*alenp = sizeof(void*);
-	} else if (ZVAL_IS_NULL(param->parameter)) {
+	} else if (ZVAL_IS_NULL(parameter)) {
 		/* insert a NULL value into the column */
 		P->indicator = -1; /* NULL */
 		*bufpp = 0;
 		*alenp = -1;
 	} else if (!P->thing) {
 		/* regular string bind */
-		convert_to_string(param->parameter);
-		*bufpp = Z_STRVAL_P(param->parameter);
-		*alenp = Z_STRLEN_P(param->parameter);
+		if (!try_convert_to_string(parameter)) {
+			return OCI_ERROR;
+		}
+		*bufpp = Z_STRVAL_P(parameter);
+		*alenp = (ub4) Z_STRLEN_P(parameter);
 	}
 
 	*piecep = OCI_ONE_PIECE;
@@ -220,11 +238,14 @@ static sb4 oci_bind_output_cb(dvoid *ctx, OCIBind *bindp, ub4 iter, ub4 index, d
 {
 	struct pdo_bound_param_data *param = (struct pdo_bound_param_data*)ctx;
 	pdo_oci_bound_param *P = (pdo_oci_bound_param*)param->driver_data;
-	TSRMLS_FETCH();
+	zval *parameter;
 
-	if (!param || !param->parameter) {
-		php_error_docref(NULL TSRMLS_CC, E_WARNING, "param is NULL in oci_bind_output_cb; this should not happen");
-		return OCI_ERROR;
+	ZEND_ASSERT(param);
+
+	if (Z_ISREF(param->parameter)) {
+		parameter = Z_REFVAL(param->parameter);
+	} else {
+		parameter = &param->parameter;
 	}
 
 	if (PDO_PARAM_TYPE(param->param_type) == PDO_PARAM_LOB) {
@@ -237,20 +258,18 @@ static sb4 oci_bind_output_cb(dvoid *ctx, OCIBind *bindp, ub4 iter, ub4 index, d
 		return OCI_CONTINUE;
 	}
 
-	if (Z_TYPE_P(param->parameter) == IS_OBJECT || Z_TYPE_P(param->parameter) == IS_RESOURCE) {
+	if (Z_TYPE_P(parameter) == IS_OBJECT || Z_TYPE_P(parameter) == IS_RESOURCE) {
 		return OCI_CONTINUE;
 	}
 
-	convert_to_string(param->parameter);
-	zval_dtor(param->parameter);
+	zval_ptr_dtor(parameter);
 
-	Z_STRLEN_P(param->parameter) = param->max_value_len;
-	Z_STRVAL_P(param->parameter) = ecalloc(1, Z_STRLEN_P(param->parameter)+1);
+	Z_STR_P(parameter) = zend_string_alloc(param->max_value_len, 1);
 	P->used_for_output = 1;
 
-	P->actual_len = Z_STRLEN_P(param->parameter);
+	P->actual_len = (ub4) Z_STRLEN_P(parameter);
 	*alenpp = &P->actual_len;
-	*bufpp = Z_STRVAL_P(param->parameter);
+	*bufpp = (Z_STR_P(parameter))->val;
 	*piecep = OCI_ONE_PIECE;
 	*rcodepp = &P->retcode;
 	*indpp = &P->indicator;
@@ -258,7 +277,7 @@ static sb4 oci_bind_output_cb(dvoid *ctx, OCIBind *bindp, ub4 iter, ub4 index, d
 	return OCI_CONTINUE;
 } /* }}} */
 
-static int oci_stmt_param_hook(pdo_stmt_t *stmt, struct pdo_bound_param_data *param, enum pdo_param_event event_type TSRMLS_DC) /* {{{ */
+static int oci_stmt_param_hook(pdo_stmt_t *stmt, struct pdo_bound_param_data *param, enum pdo_param_event event_type) /* {{{ */
 {
 	pdo_oci_stmt *S = (pdo_oci_stmt*)stmt->driver_data;
 
@@ -266,6 +285,12 @@ static int oci_stmt_param_hook(pdo_stmt_t *stmt, struct pdo_bound_param_data *pa
 	if (param->is_param) {
 		pdo_oci_bound_param *P;
 		sb4 value_sz = -1;
+		zval *parameter;
+
+		if (Z_ISREF(param->parameter))
+			parameter = Z_REFVAL(param->parameter);
+		else
+			parameter = &param->parameter;
 
 		P = (pdo_oci_bound_param*)param->driver_data;
 
@@ -278,7 +303,13 @@ static int oci_stmt_param_hook(pdo_stmt_t *stmt, struct pdo_bound_param_data *pa
 
 			case PDO_PARAM_EVT_FREE:
 				P = param->driver_data;
-				if (P) {
+				if (P && P->thing) {
+					OCI_TEMPLOB_CLOSE(S->H->env, S->H->svc, S->H->err, P->thing);
+					OCIDescriptorFree(P->thing, OCI_DTYPE_LOB);
+					P->thing = NULL;
+					efree(P);
+				}
+				else if (P) {
 					efree(P);
 				}
 				break;
@@ -295,28 +326,28 @@ static int oci_stmt_param_hook(pdo_stmt_t *stmt, struct pdo_bound_param_data *pa
 					case PDO_PARAM_LOB:
 						/* P->thing is now an OCILobLocator * */
 						P->oci_type = SQLT_BLOB;
-						value_sz = sizeof(OCILobLocator*);
+						value_sz = (sb4) sizeof(OCILobLocator*);
 						break;
 
 					case PDO_PARAM_STR:
 					default:
 						P->oci_type = SQLT_CHR;
-						value_sz = param->max_value_len;
+						value_sz = (sb4) param->max_value_len;
 						if (param->max_value_len == 0) {
-							value_sz = 1332; /* maximum size before value is interpreted as a LONG value */
+							value_sz = (sb4) 1332; /* maximum size before value is interpreted as a LONG value */
 						}
 
 				}
 
 				if (param->name) {
 					STMT_CALL(OCIBindByName, (S->stmt,
-							&P->bind, S->err, (text*)param->name,
-							param->namelen, 0, value_sz, P->oci_type,
+							&P->bind, S->err, (text*)param->name->val,
+							(sb4) param->name->len, 0, value_sz, P->oci_type,
 							&P->indicator, 0, &P->retcode, 0, 0,
 							OCI_DATA_AT_EXEC));
 				} else {
 					STMT_CALL(OCIBindByPos, (S->stmt,
-							&P->bind, S->err, param->paramno+1,
+							&P->bind, S->err, ((ub4)param->paramno)+1,
 							0, value_sz, P->oci_type,
 							&P->indicator, 0, &P->retcode, 0, 0,
 							OCI_DATA_AT_EXEC));
@@ -345,38 +376,27 @@ static int oci_stmt_param_hook(pdo_stmt_t *stmt, struct pdo_bound_param_data *pa
 				if (P->used_for_output) {
 					if (P->indicator == -1) {
 						/* set up a NULL value */
-						if (Z_TYPE_P(param->parameter) == IS_STRING
-#if ZEND_EXTENSION_API_NO < 220040718
-								&& Z_STRVAL_P(param->parameter) != empty_string
-#endif
-						   ) {
+						if (Z_TYPE_P(parameter) == IS_STRING) {
 							/* OCI likes to stick non-terminated strings in things */
-							*Z_STRVAL_P(param->parameter) = '\0';
+							*Z_STRVAL_P(parameter) = '\0';
 						}
-						zval_dtor(param->parameter);
-						ZVAL_NULL(param->parameter);
-					} else if (Z_TYPE_P(param->parameter) == IS_STRING
-#if ZEND_EXTENSION_API_NO < 220040718
-							&& Z_STRVAL_P(param->parameter) != empty_string
-#endif
-							) {
-						Z_STRLEN_P(param->parameter) = P->actual_len;
-						Z_STRVAL_P(param->parameter) = erealloc(Z_STRVAL_P(param->parameter), P->actual_len+1);
-						Z_STRVAL_P(param->parameter)[P->actual_len] = '\0';
+						zval_ptr_dtor_str(parameter);
+						ZVAL_UNDEF(parameter);
+					} else if (Z_TYPE_P(parameter) == IS_STRING) {
+						Z_STR_P(parameter) = zend_string_init(Z_STRVAL_P(parameter), P->actual_len, 1);
 					}
 				} else if (PDO_PARAM_TYPE(param->param_type) == PDO_PARAM_LOB && P->thing) {
 					php_stream *stm;
 
-					if (Z_TYPE_P(param->parameter) == IS_NULL) {
+					if (Z_TYPE_P(parameter) == IS_NULL) {
 						/* if the param is NULL, then we assume that they
 						 * wanted to bind a lob locator into it from the query
 						 * */
 
-						stm = oci_create_lob_stream(stmt, (OCILobLocator*)P->thing TSRMLS_CC);
+						stm = oci_create_lob_stream(&stmt->database_object_handle, stmt, (OCILobLocator*)P->thing);
 						if (stm) {
 							OCILobOpen(S->H->svc, S->err, (OCILobLocator*)P->thing, OCI_LOB_READWRITE);
-							php_stream_to_zval(stm, param->parameter);
-							P->thing = NULL;
+							php_stream_to_zval(stm, parameter);
 						}
 					} else {
 						/* we're a LOB being used for insert; transfer the data now */
@@ -384,7 +404,7 @@ static int oci_stmt_param_hook(pdo_stmt_t *stmt, struct pdo_bound_param_data *pa
 						ub4 amt, offset = 1;
 						char *consume;
 
-						php_stream_from_zval_no_verify(stm, &param->parameter);
+						php_stream_from_zval_no_verify(stm, parameter);
 						if (stm) {
 							OCILobOpen(S->H->svc, S->err, (OCILobLocator*)P->thing, OCI_LOB_READWRITE);
 							do {
@@ -395,9 +415,9 @@ static int oci_stmt_param_hook(pdo_stmt_t *stmt, struct pdo_bound_param_data *pa
 								}
 								consume = buf;
 								do {
-									amt = n;
+									amt = (ub4) n;
 									OCILobWrite(S->H->svc, S->err, (OCILobLocator*)P->thing,
-											&amt, offset, consume, n,
+											&amt, offset, consume, (ub4) n,
 											OCI_ONE_PIECE,
 											NULL, NULL, 0, SQLCS_IMPLICIT);
 									offset += amt;
@@ -407,16 +427,16 @@ static int oci_stmt_param_hook(pdo_stmt_t *stmt, struct pdo_bound_param_data *pa
 							} while (1);
 							OCILobClose(S->H->svc, S->err, (OCILobLocator*)P->thing);
 							OCILobFlushBuffer(S->H->svc, S->err, (OCILobLocator*)P->thing, 0);
-						} else if (Z_TYPE_P(param->parameter) == IS_STRING) {
+						} else if (Z_TYPE_P(parameter) == IS_STRING) {
 							/* stick the string into the LOB */
-							consume = Z_STRVAL_P(param->parameter);
-							n = Z_STRLEN_P(param->parameter);
+							consume = Z_STRVAL_P(parameter);
+							n = Z_STRLEN_P(parameter);
 							if (n) {
 								OCILobOpen(S->H->svc, S->err, (OCILobLocator*)P->thing, OCI_LOB_READWRITE);
 								while (n) {
-									amt = n;
+									amt = (ub4) n;
 									OCILobWrite(S->H->svc, S->err, (OCILobLocator*)P->thing,
-											&amt, offset, consume, n,
+											&amt, offset, consume, (ub4) n,
 											OCI_ONE_PIECE,
 											NULL, NULL, 0, SQLCS_IMPLICIT);
 									consume += amt;
@@ -425,6 +445,7 @@ static int oci_stmt_param_hook(pdo_stmt_t *stmt, struct pdo_bound_param_data *pa
 								OCILobClose(S->H->svc, S->err, (OCILobLocator*)P->thing);
 							}
 						}
+						OCI_TEMPLOB_CLOSE(S->H->env, S->H->svc, S->H->err, P->thing);
 						OCIDescriptorFree(P->thing, OCI_DTYPE_LOB);
 						P->thing = NULL;
 					}
@@ -437,14 +458,14 @@ static int oci_stmt_param_hook(pdo_stmt_t *stmt, struct pdo_bound_param_data *pa
 	return 1;
 } /* }}} */
 
-static int oci_stmt_fetch(pdo_stmt_t *stmt, enum pdo_fetch_orientation ori,	long offset TSRMLS_DC) /* {{{ */
+static int oci_stmt_fetch(pdo_stmt_t *stmt, enum pdo_fetch_orientation ori,	zend_long offset) /* {{{ */
 {
-#if HAVE_OCISTMTFETCH2
-	ub4 ociori;
+#ifdef HAVE_OCISTMTFETCH2
+	ub4 ociori = OCI_FETCH_NEXT;
 #endif
 	pdo_oci_stmt *S = (pdo_oci_stmt*)stmt->driver_data;
 
-#if HAVE_OCISTMTFETCH2
+#ifdef HAVE_OCISTMTFETCH2
 	switch (ori) {
 		case PDO_FETCH_ORI_NEXT:	ociori = OCI_FETCH_NEXT; break;
 		case PDO_FETCH_ORI_PRIOR:	ociori = OCI_FETCH_PRIOR; break;
@@ -453,7 +474,7 @@ static int oci_stmt_fetch(pdo_stmt_t *stmt, enum pdo_fetch_orientation ori,	long
 		case PDO_FETCH_ORI_ABS:		ociori = OCI_FETCH_ABSOLUTE; break;
 		case PDO_FETCH_ORI_REL:		ociori = OCI_FETCH_RELATIVE; break;
 	}
-	S->last_err = OCIStmtFetch2(S->stmt, S->err, 1, ociori, offset, OCI_DEFAULT);
+	S->last_err = OCIStmtFetch2(S->stmt, S->err, 1, ociori, (sb4) offset, OCI_DEFAULT);
 #else
 	S->last_err = OCIStmtFetch(S->stmt, S->err, 1, OCI_FETCH_NEXT, OCI_DEFAULT);
 #endif
@@ -481,7 +502,6 @@ static sb4 oci_define_callback(dvoid *octxp, OCIDefine *define, ub4 iter, dvoid
 		ub4 **alenpp, ub1 *piecep, dvoid **indpp, ub2 **rcodepp)
 {
 	pdo_oci_column *col = (pdo_oci_column*)octxp;
-	TSRMLS_FETCH();
 
 	switch (col->dtype) {
 		case SQLT_BLOB:
@@ -491,25 +511,21 @@ static sb4 oci_define_callback(dvoid *octxp, OCIDefine *define, ub4 iter, dvoid
 			*alenpp = &col->datalen;
 			*indpp = (dvoid *)&col->indicator;
 			break;
-
-		default:
-			php_error_docref(NULL TSRMLS_CC, E_WARNING,
-				"unhandled datatype in oci_define_callback; this should not happen");
-			return OCI_ERROR;
+		EMPTY_SWITCH_DEFAULT_CASE();
 	}
 
 	return OCI_CONTINUE;
 }
 
-static int oci_stmt_describe(pdo_stmt_t *stmt, int colno TSRMLS_DC) /* {{{ */
+static int oci_stmt_describe(pdo_stmt_t *stmt, int colno) /* {{{ */
 {
 	pdo_oci_stmt *S = (pdo_oci_stmt*)stmt->driver_data;
 	OCIParam *param = NULL;
 	text *colname;
-	ub2 dtype, data_size, scale, precis;
+	ub2 dtype, data_size, precis;
 	ub4 namelen;
 	struct pdo_column_data *col = &stmt->columns[colno];
-	zend_bool dyn = FALSE;
+	bool dyn = FALSE;
 
 	/* describe the column */
 	STMT_CALL(OCIParamGet, (S->stmt, OCI_HTYPE_STMT, S->err, (dvoid*)&param, colno+1));
@@ -522,10 +538,6 @@ static int oci_stmt_describe(pdo_stmt_t *stmt, int colno TSRMLS_DC) /* {{{ */
 	STMT_CALL_MSG(OCIAttrGet, "OCI_ATTR_DATA_SIZE",
 			(param, OCI_DTYPE_PARAM, &data_size, 0, OCI_ATTR_DATA_SIZE, S->err));
 
-	/* scale ? */
-	STMT_CALL_MSG(OCIAttrGet, "OCI_ATTR_SCALE",
-			(param, OCI_DTYPE_PARAM, &scale, 0, OCI_ATTR_SCALE, S->err));
-
 	/* precision ? */
 	STMT_CALL_MSG(OCIAttrGet, "OCI_ATTR_PRECISION",
 			(param, OCI_DTYPE_PARAM, &precis, 0, OCI_ATTR_PRECISION, S->err));
@@ -534,10 +546,9 @@ static int oci_stmt_describe(pdo_stmt_t *stmt, int colno TSRMLS_DC) /* {{{ */
 	STMT_CALL_MSG(OCIAttrGet, "OCI_ATTR_NAME",
 			(param, OCI_DTYPE_PARAM, &colname, &namelen, OCI_ATTR_NAME, S->err));
 
-	col->precision = scale;
+	col->precision = precis;
 	col->maxlen = data_size;
-	col->namelen = namelen;
-	col->name = estrndup((char *)colname, namelen);
+	col->name = zend_string_init((char *)colname, namelen, 0);
 
 	S->cols[colno].dtype = dtype;
 
@@ -552,12 +563,10 @@ static int oci_stmt_describe(pdo_stmt_t *stmt, int colno TSRMLS_DC) /* {{{ */
 			}
 			S->cols[colno].datalen = 512; /* XXX should be INT_MAX and fetched by pieces */
 			S->cols[colno].data = emalloc(S->cols[colno].datalen + 1);
-			col->param_type = PDO_PARAM_STR;
 			break;
 
 		case SQLT_BLOB:
 		case SQLT_CLOB:
-			col->param_type = PDO_PARAM_LOB;
 			STMT_CALL(OCIDescriptorAlloc, (S->H->env, (dvoid**)&S->cols[colno].data, OCI_DTYPE_LOB, 0, NULL));
 			S->cols[colno].datalen = sizeof(OCILobLocator*);
 			dyn = TRUE;
@@ -579,17 +588,14 @@ static int oci_stmt_describe(pdo_stmt_t *stmt, int colno TSRMLS_DC) /* {{{ */
 			} else if (dtype == SQLT_IBFLOAT || dtype == SQLT_IBDOUBLE) {
 				S->cols[colno].datalen = 1024;
 #endif
+			} else if (dtype == SQLT_BIN) {
+				S->cols[colno].datalen = (ub4) col->maxlen * 2; /* raw characters to hex digits */
 			} else {
-				S->cols[colno].datalen = col->maxlen;
-			}
-			if (dtype == SQLT_BIN) {
-				S->cols[colno].datalen *= 3;
+				S->cols[colno].datalen = (ub4) (col->maxlen * S->H->max_char_width);
 			}
+
 			S->cols[colno].data = emalloc(S->cols[colno].datalen + 1);
 			dtype = SQLT_CHR;
-
-			/* returning data as a string */
-			col->param_type = PDO_PARAM_STR;
 	}
 
 	STMT_CALL(OCIDefineByPos, (S->stmt, &S->cols[colno].def, S->err, colno+1,
@@ -604,89 +610,114 @@ static int oci_stmt_describe(pdo_stmt_t *stmt, int colno TSRMLS_DC) /* {{{ */
 	return 1;
 } /* }}} */
 
+struct _oci_lob_env {
+	OCISvcCtx *svc;
+	OCIError  *err;
+};
+typedef struct _oci_lob_env oci_lob_env;
+
 struct oci_lob_self {
+	zval dbh;
 	pdo_stmt_t *stmt;
 	pdo_oci_stmt *S;
 	OCILobLocator *lob;
+	oci_lob_env   *E;
 	ub4 offset;
+	ub1 csfrm;
 };
 
-static size_t oci_blob_write(php_stream *stream, const char *buf, size_t count TSRMLS_DC)
+static ssize_t oci_blob_write(php_stream *stream, const char *buf, size_t count)
 {
 	struct oci_lob_self *self = (struct oci_lob_self*)stream->abstract;
 	ub4 amt;
 	sword r;
 
-	amt = count;
-	r = OCILobWrite(self->S->H->svc, self->S->err, self->lob,
-		&amt, self->offset, (char*)buf, count,
+	amt = (ub4) count;
+	r = OCILobWrite(self->E->svc, self->E->err, self->lob,
+		&amt, self->offset, (char*)buf, (ub4) count,
 		OCI_ONE_PIECE,
 		NULL, NULL, 0, SQLCS_IMPLICIT);
 
 	if (r != OCI_SUCCESS) {
-		return (size_t)-1;
+		return (ssize_t)-1;
 	}
 
 	self->offset += amt;
 	return amt;
 }
 
-static size_t oci_blob_read(php_stream *stream, char *buf, size_t count TSRMLS_DC)
+static ssize_t oci_blob_read(php_stream *stream, char *buf, size_t count)
 {
 	struct oci_lob_self *self = (struct oci_lob_self*)stream->abstract;
-	ub4 amt;
-	sword r;
+#if HAVE_OCILOBREAD2
+	oraub8 byte_amt = (oraub8) count;
+	oraub8 char_amt = 0;
 
-	amt = count;
-	r = OCILobRead(self->S->H->svc, self->S->err, self->lob,
-		&amt, self->offset, buf, count,
+	sword r = OCILobRead2(self->E->svc, self->E->err, self->lob,
+		&byte_amt, &char_amt, (oraub8) self->offset, buf, (oraub8) count,
+        OCI_ONE_PIECE, NULL, NULL, 0, self->csfrm);
+#else
+	ub4 byte_amt = (ub4) count;
+
+	sword r = OCILobRead(self->E->svc, self->E->err, self->lob,
+		&byte_amt, self->offset, buf, (ub4) count,
 		NULL, NULL, 0, SQLCS_IMPLICIT);
+#endif
 
 	if (r != OCI_SUCCESS && r != OCI_NEED_DATA) {
-		return (size_t)-1;
+		return (ssize_t)-1;
 	}
 
-	self->offset += amt;
-	if (amt < count) {
+#if HAVE_OCILOBREAD2
+	self->offset += self->csfrm == 0 ? byte_amt : char_amt;
+#else
+	self->offset += byte_amt;
+#endif
+	if (byte_amt < count) {
 		stream->eof = 1;
 	}
-	return amt;
+	return byte_amt;
 }
 
-static int oci_blob_close(php_stream *stream, int close_handle TSRMLS_DC)
+static int oci_blob_close(php_stream *stream, int close_handle)
 {
-	struct oci_lob_self *self = (struct oci_lob_self*)stream->abstract;
+	struct oci_lob_self *self = (struct oci_lob_self *)stream->abstract;
 	pdo_stmt_t *stmt = self->stmt;
 
 	if (close_handle) {
-		OCILobClose(self->S->H->svc, self->S->err, self->lob);
+		zend_object *obj = &stmt->std;
+
+		OCILobClose(self->E->svc, self->E->err, self->lob);
+		zval_ptr_dtor(&self->dbh);
+		GC_DELREF(obj);
+		efree(self->E);
 		efree(self);
 	}
 
-	php_pdo_stmt_delref(stmt TSRMLS_CC);
+	/* php_pdo_free_statement(stmt); */
 	return 0;
 }
 
-static int oci_blob_flush(php_stream *stream TSRMLS_DC)
+static int oci_blob_flush(php_stream *stream)
 {
 	struct oci_lob_self *self = (struct oci_lob_self*)stream->abstract;
-	OCILobFlushBuffer(self->S->H->svc, self->S->err, self->lob, 0);
+	OCILobFlushBuffer(self->E->svc, self->E->err, self->lob, 0);
 	return 0;
 }
 
-static int oci_blob_seek(php_stream *stream, off_t offset, int whence, off_t *newoffset TSRMLS_DC)
+static int oci_blob_seek(php_stream *stream, zend_off_t offset, int whence, zend_off_t *newoffset)
 {
 	struct oci_lob_self *self = (struct oci_lob_self*)stream->abstract;
 
 	if (offset >= PDO_OCI_LOBMAXSIZE) {
 		return -1;
 	} else {
-		self->offset = offset + 1;  /* Oracle LOBS are 1-based, but PHP is 0-based */
+		self->offset = (ub4) offset + 1;  /* Oracle LOBS are 1-based, but PHP is 0-based */
 		return 0;
 	}
 }
 
-static php_stream_ops oci_blob_stream_ops = {
+static const php_stream_ops oci_blob_stream_ops = {
 	oci_blob_write,
 	oci_blob_read,
 	oci_blob_close,
@@ -698,19 +729,29 @@ static php_stream_ops oci_blob_stream_ops = {
 	NULL
 };
 
-static php_stream *oci_create_lob_stream(pdo_stmt_t *stmt, OCILobLocator *lob TSRMLS_DC)
+static php_stream *oci_create_lob_stream(zval *dbh, pdo_stmt_t *stmt, OCILobLocator *lob)
 {
 	php_stream *stm;
 	struct oci_lob_self *self = ecalloc(1, sizeof(*self));
+
+	ZVAL_COPY_VALUE(&self->dbh, dbh);
 	self->lob = lob;
 	self->offset = 1; /* 1-based */
 	self->stmt = stmt;
 	self->S = (pdo_oci_stmt*)stmt->driver_data;
+	self->E = ecalloc(1, sizeof(oci_lob_env));
+	self->E->svc = self->S->H->svc;
+	self->E->err = self->S->err;
+
+	OCILobCharSetForm(self->S->H->env, self->S->err, self->lob, &self->csfrm);
 
 	stm = php_stream_alloc(&oci_blob_stream_ops, self, 0, "r+b");
 
 	if (stm) {
-		php_pdo_stmt_addref(stmt TSRMLS_CC);
+		zend_object *obj;
+		obj = &stmt->std;
+		Z_ADDREF(self->dbh);
+		GC_ADDREF(obj);
 		return stm;
 	}
 
@@ -718,7 +759,7 @@ static php_stream *oci_create_lob_stream(pdo_stmt_t *stmt, OCILobLocator *lob TS
 	return NULL;
 }
 
-static int oci_stmt_get_col(pdo_stmt_t *stmt, int colno, char **ptr, unsigned long *len, int *caller_frees TSRMLS_DC) /* {{{ */
+static int oci_stmt_get_col(pdo_stmt_t *stmt, int colno, zval *result, enum pdo_param_type *type) /* {{{ */
 {
 	pdo_oci_stmt *S = (pdo_oci_stmt*)stmt->driver_data;
 	pdo_oci_column *C = &S->cols[colno];
@@ -726,48 +767,244 @@ static int oci_stmt_get_col(pdo_stmt_t *stmt, int colno, char **ptr, unsigned lo
 	/* check the indicator to ensure that the data is intact */
 	if (C->indicator == -1) {
 		/* A NULL value */
-		*ptr = NULL;
-		*len = 0;
+		ZVAL_NULL(result);
 		return 1;
 	} else if (C->indicator == 0) {
 		/* it was stored perfectly */
 
 		if (C->dtype == SQLT_BLOB || C->dtype == SQLT_CLOB) {
 			if (C->data) {
-				*ptr = (char*)oci_create_lob_stream(stmt, (OCILobLocator*)C->data TSRMLS_CC);
+				php_stream *stream = oci_create_lob_stream(&stmt->database_object_handle, stmt, (OCILobLocator*)C->data);
 				OCILobOpen(S->H->svc, S->err, (OCILobLocator*)C->data, OCI_LOB_READONLY);
+				php_stream_to_zval(stream, result);
+				return 1;
 			}
-			*len = 0;
-			return *ptr ? 1 : 0;
+			return 0;
 		}
 
-		*ptr = C->data;
-		*len = C->fetched_len;
+		ZVAL_STRINGL_FAST(result, C->data, C->fetched_len);
 		return 1;
 	} else {
 		/* it was truncated */
-		php_error_docref(NULL TSRMLS_CC, E_WARNING, "column %d data was too large for buffer and was truncated to fit it", colno);
-
-		*ptr = C->data;
-		*len = C->fetched_len;
+		php_error_docref(NULL, E_WARNING, "Column %d data was too large for buffer and was truncated to fit it", colno);
+		ZVAL_STRINGL(result, C->data, C->fetched_len);
 		return 1;
 	}
 } /* }}} */
 
-struct pdo_stmt_methods oci_stmt_methods = {
+
+static int oci_stmt_col_meta(pdo_stmt_t *stmt, zend_long colno, zval *return_value) /* {{{ */
+{
+	pdo_oci_stmt *S = (pdo_oci_stmt*)stmt->driver_data;
+	OCIParam *param = NULL;
+	ub2 dtype, precis;
+	sb1 scale;
+	zval flags;
+	ub1 isnull, charset_form;
+	if (!S->stmt) {
+		return FAILURE;
+	}
+	if (colno >= stmt->column_count) {
+		/* error invalid column */
+		return FAILURE;
+	}
+
+	array_init(return_value);
+	array_init(&flags);
+
+	/* describe the column */
+	STMT_CALL(OCIParamGet, (S->stmt, OCI_HTYPE_STMT, S->err, (dvoid*)&param, colno+1));
+
+	/* column data type */
+	STMT_CALL_MSG(OCIAttrGet, "OCI_ATTR_DATA_TYPE",
+			(param, OCI_DTYPE_PARAM, &dtype, 0, OCI_ATTR_DATA_TYPE, S->err));
+
+	/* column precision */
+	STMT_CALL_MSG(OCIAttrGet, "OCI_ATTR_PRECISION",
+			(param, OCI_DTYPE_PARAM, &precis, 0, OCI_ATTR_PRECISION, S->err));
+
+	/* column scale */
+	STMT_CALL_MSG(OCIAttrGet, "OCI_ATTR_SCALE",
+			(param, OCI_DTYPE_PARAM, &scale, 0, OCI_ATTR_SCALE, S->err));
+
+	/* string column charset form */
+	if (dtype == SQLT_CHR || dtype == SQLT_VCS || dtype == SQLT_AFC || dtype == SQLT_CLOB) {
+		STMT_CALL_MSG(OCIAttrGet, "OCI_ATTR_CHARSET_FORM",
+			(param, OCI_DTYPE_PARAM, &charset_form, 0, OCI_ATTR_CHARSET_FORM, S->err));
+	}
+
+
+	if (dtype) {
+	/* if there is a declared type */
+		switch (dtype) {
+#ifdef SQLT_TIMESTAMP
+		case SQLT_TIMESTAMP:
+			add_assoc_string(return_value, "oci:decl_type", "TIMESTAMP");
+			add_assoc_string(return_value, "native_type", "TIMESTAMP");
+			break;
+#endif
+#ifdef SQLT_TIMESTAMP_TZ
+		case SQLT_TIMESTAMP_TZ:
+			add_assoc_string(return_value, "oci:decl_type", "TIMESTAMP WITH TIMEZONE");
+			add_assoc_string(return_value, "native_type", "TIMESTAMP WITH TIMEZONE");
+			break;
+#endif
+#ifdef SQLT_TIMESTAMP_LTZ
+		case SQLT_TIMESTAMP_LTZ:
+			add_assoc_string(return_value, "oci:decl_type", "TIMESTAMP WITH LOCAL TIMEZONE");
+			add_assoc_string(return_value, "native_type", "TIMESTAMP WITH LOCAL TIMEZONE");
+			break;
+#endif
+#ifdef SQLT_INTERVAL_YM
+		case SQLT_INTERVAL_YM:
+			add_assoc_string(return_value, "oci:decl_type", "INTERVAL YEAR TO MONTH");
+			add_assoc_string(return_value, "native_type", "INTERVAL YEAR TO MONTH");
+			break;
+#endif
+#ifdef SQLT_INTERVAL_DS
+		case SQLT_INTERVAL_DS:
+			add_assoc_string(return_value, "oci:decl_type", "INTERVAL DAY TO SECOND");
+			add_assoc_string(return_value, "native_type", "INTERVAL DAY TO SECOND");
+			break;
+#endif
+		case SQLT_DAT:
+			add_assoc_string(return_value, "oci:decl_type", "DATE");
+			add_assoc_string(return_value, "native_type", "DATE");
+			break;
+		case SQLT_FLT :
+		case SQLT_NUM:
+			/* if the precision is nonzero and scale is -127 then it is a FLOAT */
+			if (scale == -127 && precis != 0) {
+				add_assoc_string(return_value, "oci:decl_type", "FLOAT");
+				add_assoc_string(return_value, "native_type", "FLOAT");
+			} else {
+				add_assoc_string(return_value, "oci:decl_type", "NUMBER");
+				add_assoc_string(return_value, "native_type", "NUMBER");
+			}
+			break;
+		case SQLT_LNG:
+			add_assoc_string(return_value, "oci:decl_type", "LONG");
+			add_assoc_string(return_value, "native_type", "LONG");
+			break;
+		case SQLT_BIN:
+			add_assoc_string(return_value, "oci:decl_type", "RAW");
+			add_assoc_string(return_value, "native_type", "RAW");
+			break;
+		case SQLT_LBI:
+			add_assoc_string(return_value, "oci:decl_type", "LONG RAW");
+			add_assoc_string(return_value, "native_type", "LONG RAW");
+			break;
+		case SQLT_CHR:
+		case SQLT_VCS:
+			if (charset_form == SQLCS_NCHAR) {
+				add_assoc_string(return_value, "oci:decl_type", "NVARCHAR2");
+				add_assoc_string(return_value, "native_type", "NVARCHAR2");
+			} else {
+				add_assoc_string(return_value, "oci:decl_type", "VARCHAR2");
+				add_assoc_string(return_value, "native_type", "VARCHAR2");
+			}
+			break;
+		case SQLT_AFC:
+			if (charset_form == SQLCS_NCHAR) {
+				add_assoc_string(return_value, "oci:decl_type", "NCHAR");
+				add_assoc_string(return_value, "native_type", "NCHAR");
+			} else {
+				add_assoc_string(return_value, "oci:decl_type", "CHAR");
+				add_assoc_string(return_value, "native_type", "CHAR");
+			}
+			break;
+		case SQLT_BLOB:
+			add_assoc_string(return_value, "oci:decl_type", "BLOB");
+			add_next_index_string(&flags, "blob");
+			add_assoc_string(return_value, "native_type", "BLOB");
+			break;
+		case SQLT_CLOB:
+			if (charset_form == SQLCS_NCHAR) {
+				add_assoc_string(return_value, "oci:decl_type", "NCLOB");
+				add_assoc_string(return_value, "native_type", "NCLOB");
+			} else {
+				add_assoc_string(return_value, "oci:decl_type", "CLOB");
+				add_assoc_string(return_value, "native_type", "CLOB");
+			}
+			add_next_index_string(&flags, "blob");
+			break;
+		case SQLT_BFILE:
+			add_assoc_string(return_value, "oci:decl_type", "BFILE");
+			add_next_index_string(&flags, "blob");
+			add_assoc_string(return_value, "native_type", "BFILE");
+			break;
+		case SQLT_RDD:
+			add_assoc_string(return_value, "oci:decl_type", "ROWID");
+			add_assoc_string(return_value, "native_type", "ROWID");
+			break;
+		case SQLT_BFLOAT:
+		case SQLT_IBFLOAT:
+			add_assoc_string(return_value, "oci:decl_type", "BINARY_FLOAT");
+			add_assoc_string(return_value, "native_type", "BINARY_FLOAT");
+			break;
+		case SQLT_BDOUBLE:
+		case SQLT_IBDOUBLE:
+			add_assoc_string(return_value, "oci:decl_type", "BINARY_DOUBLE");
+			add_assoc_string(return_value, "native_type", "BINARY_DOUBLE");
+			break;
+		default:
+			add_assoc_long(return_value, "oci:decl_type", dtype);
+			add_assoc_string(return_value, "native_type", "UNKNOWN");
+		}
+	} else {
+		/* if the column is NULL */
+		add_assoc_long(return_value, "oci:decl_type", 0);
+		add_assoc_string(return_value, "native_type", "NULL");
+	}
+
+	switch (dtype) {
+		case SQLT_BLOB:
+		case SQLT_CLOB:
+			add_assoc_long(return_value, "pdo_type", PDO_PARAM_LOB);
+			break;
+		default:
+			add_assoc_long(return_value, "pdo_type", PDO_PARAM_STR);
+			break;
+	}
+
+	/* column can be null */
+	STMT_CALL_MSG(OCIAttrGet, "OCI_ATTR_IS_NULL",
+			(param, OCI_DTYPE_PARAM, &isnull, 0, OCI_ATTR_IS_NULL, S->err));
+
+	if (isnull) {
+		add_next_index_string(&flags, "nullable");
+	} else {
+		add_next_index_string(&flags, "not_null");
+	}
+
+	/* PDO type */
+	switch (dtype) {
+		case SQLT_BFILE:
+		case SQLT_BLOB:
+		case SQLT_CLOB:
+			add_assoc_long(return_value, "pdo_type", PDO_PARAM_LOB);
+			break;
+		default:
+			add_assoc_long(return_value, "pdo_type", PDO_PARAM_STR);
+	}
+
+	add_assoc_long(return_value, "scale", scale);
+	add_assoc_zval(return_value, "flags", &flags);
+
+	OCIDescriptorFree(param, OCI_DTYPE_PARAM);
+	return SUCCESS;
+} /* }}} */
+
+const struct pdo_stmt_methods oci_stmt_methods = {
 	oci_stmt_dtor,
 	oci_stmt_execute,
 	oci_stmt_fetch,
 	oci_stmt_describe,
 	oci_stmt_get_col,
-	oci_stmt_param_hook
+	oci_stmt_param_hook,
+	NULL, /* set_attr */
+	NULL, /* get_attr */
+	oci_stmt_col_meta,
+	NULL,
+	NULL
 };
-
-/*
- * Local variables:
- * tab-width: 4
- * c-basic-offset: 4
- * End:
- * vim600: noet sw=4 ts=4 fdm=marker
- * vim<600: noet sw=4 ts=4
- */
